require('dotenv').config(); // Load environment variables
const { exec } = require('child_process');
const dgram = require('dgram');
const dnsPacket = require('dns-packet');
const HolesailClient = require('holesail-client');
const Corestore = require('corestore');
const Hyperswarm = require('hyperswarm');
const http = require('http');
const b4a = require('b4a');
const { createHash } = require('crypto');
const net = require('net');

// Corestore and Hyperswarm setup for P2P sync
const store = new Corestore('./my-storage');
const swarm = new Hyperswarm();

// Option for setting a customizable masterNetwork_discoveryKey for TLD syncing
const masterNetworkDiscoveryKey = process.env.masterNetworkDiscoveryKey 
    ? Buffer.from(process.env.masterNetworkDiscoveryKey, 'hex') 
    : null;

const dnsCore = store.get({ name: 'dns-core' });

let holesailClient = null;
const dnsServer = dgram.createSocket('udp4');
const domainToIPMap = {};
let currentIP = 2; // Start assigning IP addresses from 192.168.100.2

console.log(`Loaded masterNetworkDiscoveryKey: ${masterNetworkDiscoveryKey ? masterNetworkDiscoveryKey.toString('hex') : 'None'}`);

// Helper function to remove existing virtual interface if it already exists (for macOS)
function removeExistingInterface(subnetName) {
    return new Promise((resolve) => {
        exec(`sudo ifconfig ${subnetName} down`, (err) => {
            if (err) resolve();
            else {
                console.log(`Removed existing virtual interface: ${subnetName}`);
                resolve();
            }
        });
    });
}

// Helper function to create virtual interfaces (for macOS)
async function createVirtualInterface(subnetName, subnetCIDR) {
    await removeExistingInterface(subnetName);
    return new Promise((resolve, reject) => {
        exec(`sudo ifconfig ${subnetName} alias ${subnetCIDR}`, (err, stdout, stderr) => {
            if (err) {
                console.error(`Error creating virtual interface ${subnetName}:`, stderr);
                reject(`Error creating virtual interface ${subnetName}: ${stderr}`);
            } else {
                console.log(`Virtual interface ${subnetName} created with CIDR ${subnetCIDR}.`);
                resolve(subnetCIDR);
            }
        });
    });
}

// Updated function to create virtual interfaces dynamically, starting from 192.168.100.2
async function createInterfaceForDomain(domain) {
    const subnetID = currentIP;
    const subnetName = `lo0`;
    const subnetCIDR = `192.168.100.${subnetID}/24`;

    try {
        await createVirtualInterface(subnetName, subnetCIDR);
        domainToIPMap[domain] = `192.168.100.${subnetID}`;
        logDebug(`Assigned virtual interface IP: 192.168.100.${subnetID} for domain: ${domain}`);
        currentIP++;
        return domainToIPMap[domain];
    } catch (error) {
        console.error(error);
        return null;
    }
}

// Debug log wrapper for standardized output
function logDebug(info) {
    const timestamp = new Date().toISOString();
    console.log(`[DEBUG ${timestamp}] ${info}`);
}

// Function to check if a port is responsive
function checkPortResponsive(ip, port) {
    return new Promise((resolve) => {
        const client = new net.Socket();
        client.setTimeout(2000);

        client.connect(port, ip, () => {
            client.destroy();
            resolve(true);
        });

        client.on('error', () => resolve(false));
        client.on('timeout', () => {
            client.destroy();
            resolve(false);
        });
    });
}

// Function to check public DNS using Cloudflare
function checkPublicDNS(domain) {
    return new Promise((resolve) => {
        const resolver = dgram.createSocket('udp4');
        const query = dnsPacket.encode({
            type: 'query',
            id: Math.floor(Math.random() * 65535),
            questions: [{ type: 'A', name: domain }]
        });

        resolver.send(query, 53, '1.1.1.1', (err) => {
            if (err) {
                logDebug(`Error forwarding DNS query to 1.1.1.1: ${err}`);
                return resolve(null);
            }
        });

        resolver.on('message', (res) => {
            const response = dnsPacket.decode(res);
            if (response.answers.length > 0) {
                logDebug(`Public DNS returned records for ${domain}`);
                resolve(response.answers);
            } else {
                resolve(null);
            }
            resolver.close();
        });
    });
}

const activeConnections = {}; // Store active connections by domain

// Function to start Holesail client for tunneling and use port 80 with SO_REUSEADDR
function startHolesailClient(domain, hash, ip, port) {
    logDebug(`Attempting to start/reuse Holesail client for domain: ${domain}`);

    if (activeConnections[domain]) {
        logDebug(`Reusing existing Holesail client for domain: ${domain} on ${ip}:${port}`);
        return activeConnections[domain]; // Reuse the existing connection
    }

    logDebug(`Starting new Holesail client for domain: ${domain}, hash: ${hash}, IP: ${ip}, Port: ${port}`);

    const connector = setupConnector(hash);
    const holesailClient = new HolesailClient(connector);

    holesailClient.connect({ port: port, address: ip, reuseAddr: true }, () => {
        logDebug(`Holesail client for ${domain} connected on ${ip}:${port}`);
    });

    // Store the client in activeConnections
    activeConnections[domain] = holesailClient;

    // Set a timeout to destroy and remove from activeConnections after 5 minutes (300,000 ms)
    setTimeout(() => {
        logDebug(`Destroying Holesail client for domain ${domain}`);
        holesailClient.destroy();
        delete activeConnections[domain]; // Remove the client from activeConnections
    }, 300000);

    return holesailClient;
}

// Ensure the DNS Core is ready before joining the swarm
(async () => {
    await dnsCore.ready();

    // Use masterNetworkDiscoveryKey if provided; otherwise, use dnsCore.discoveryKey
    const topic = masterNetworkDiscoveryKey || dnsCore.discoveryKey;

    logDebug(`DNS Core ready, joining Hyperswarm with topic: ${topic.toString('hex')}`);

    swarm.join(topic, { server: true, client: true });

    swarm.on('connection', (conn) => {
        logDebug('Peer connected, starting replication...');
        dnsCore.replicate(conn);
    });
})();

// Function to add a domain and hash to the DNS core (P2P network)
async function addDomain(domain, hash) {
    await dnsCore.ready();
    const record = JSON.stringify({ domain, hash });

    logDebug(`Adding domain ${domain} with hash ${hash} to DNS core`);
    await dnsCore.append(Buffer.from(record));
    logDebug(`Domain ${domain} added to DNS core`);
}

// Example: Add a domain to the P2P DNS system
addDomain('hello.geek', '07b8b52fbbad7a89ce26ad2d8375e6a82b2e3c02f18596bddff18e9c31164b04');
